# Code used in lectures of
# ST414: Advanced topics in Statistics
# Asymptotic Statistics
###
# 4th December 2009
# Ioannis Kosmidis, http://go.warwick.ac.uk/kosmidis

set.seed(10)
truenu <- 1
truelambda <- 1
ssize <- 20
y <- rweibull(ssize, shape = truenu, scale = truelambda^(-1/truenu))

## the Weibull log-likelihood
weibull.loglik <- function(beta, dat) {
  lambda <- beta[1]
  nu <- beta[2]
  sum(dweibull(dat, shape = nu,
                scale = lambda^(-1/nu), log = TRUE))
}

## the score function
weibull.score <- function(beta, dat) {
  n <- length(dat)
  nu <- beta[2]
  lambda <- beta[1]
  -c(-sum(dat^nu) + n/lambda,
     n/nu + sum(log(dat)) - lambda*sum(log(dat)*dat^nu))
}

## the observed information
weibull.info <- function(beta, dat) {
  n <- length(dat)
  lambda <- beta[1]
  nu <- beta[2]
  jll <- n/lambda^2
  jln <- sum(dat^nu*log(dat))
  jnn <- n/nu^2 + lambda*sum(dat^nu*(log(dat))^2)
  matrix(c(jll, jln, jln, jnn), nrow = 2, ncol =2)
}

## Maximum likelihood estimates
nn <- length(y)
nuhat <- uniroot(f = function(nu1) nn/nu1 + sum(log(y)) -
                 nn*sum(y^nu1*log(y))/sum(y^nu1),
                 interval = c(0.001, 10), tol = 1e-8)$root
lambdahat <- nn/sum(y^nuhat)
mles <- c(lambdahat, nuhat)
names(mles) <- c("lambdahat","nuhat")
## You can check that this is the solution by
# weibull.score(mles, dat = y)
#
## Estimated asymptotic standard errors for the MLE
jhat <- weibull.info(mles, dat = y)
jhatInverse <- solve(jhat)
mle.std.errs <- sqrt(diag(jhatInverse))


## below we set a grid. The endpoints might need to be changed
## for different seed or different sample size (original 20)
lambdas <- seq(0.01, 3, length = 200)
nus <- seq(0.01, 3, length = 200)
parvalues <- expand.grid(lambdas, nus)
##
## levels to be plotted, colors, line widths
conf.levels <- c(0.001, 0.5, 0.7, 0.9, 0.95, 0.99)
if (!(0.95%in%conf.levels))
  conf.levels <- c(conf.levels, 0.95)
color1 <- rep("black", length(conf.levels))
lwds <- rep(1, length(conf.levels))
lwds[which(conf.levels == 0.95)] <- 2
#color1[which(conf.levels == 0.95)] <- "grey"

##
##
## Confidence regions based on w(lambda, nu)
dev.new()
llikvalues <- apply(parvalues, 1, weibull.loglik, dat = y)
llikvalues <- matrix(llikvalues,
                     nrow = length(lambdas),
                     ncol = length(nus))
contour(lambdas, nus, 2*(max(llikvalues) - llikvalues),
        levels = qchisq(conf.levels, 2),
        labels = as.character(conf.levels),
        xlab = expression(lambda),
        ylab = expression(nu),
        col = color1, lwd = lwds)
points(x = mles[1], y = mles[2], col = "red", pch =16)
title(expression(w(lambda, nu)))

##
##
## Confidence regions based on s(lambda, nu)
##
dev.new()
sSqvalues <- apply(parvalues, 1,
                   function(pars, dat) {
                     ss <- as.matrix(weibull.score(pars, dat))
                     c(t(ss)%*%jhatInverse%*%ss)
                   }, dat = y)
sSqvalues <- matrix(sSqvalues,
                    nrow = length(lambdas),
                    ncol = length(nus))
contour(lambdas, nus, sSqvalues,
        levels = qchisq(conf.levels, 2),
        labels = as.character(conf.levels),
        xlab = expression(lambda),
        ylab = expression(nu),
        col = color1, lwd = lwds)
points(x = mles[1], y = mles[2], col = "red", pch =16)
title(expression({s^2}(lambda, nu)))

##
##
## Confidence regions based on t^2(lambda, nu)
##
dev.new()
tSqvalues <- apply(parvalues, 1,
                   function(pars, dat) {
                     diff1 <- as.matrix(pars - mles)
                     c(t(diff1)%*%jhat%*%diff1)
                   }, dat = y)
tSqvalues <- matrix(tSqvalues,
                    nrow = length(lambdas),
                    ncol = length(nus))
contour(lambdas, nus, tSqvalues,
        levels = qchisq(conf.levels, 2),
        labels = as.character(conf.levels),
        xlab = expression(lambda),
        ylab = expression(nu),
        col = color1, lwd = lwds)
points(x = mles[1], y = mles[2], col = "red", pch =16)
title(expression({t^2}(lambda, nu)))


####
### Estimate the coverage of confidence regions 
### at nomial levels conf.levels
### based on a simulation of size nsimu
### for sample size ssize
### and true parameter vector beta0
weibull.simu <- function(nsimu, ssize, beta0 = c(1, 1),
                         conf.levels = c(0.9, 0.95, 0.99)) {
  values1 <- function(dat) {
    nn <- length(dat)
    nuhat <- uniroot(f = function(nu1) nn/nu1 + sum(log(dat)) -
                     nn*sum(dat^nu1*log(dat))/sum(dat^nu1),
                     interval = c(0.0001, 10), tol = 1e-8)$root
    lambdahat <- nn/sum(dat^nuhat)
    mles <- c(lambdahat, nuhat)
    #ml.fit <- optim(par = c(1, 1), 
    #                fn = function(x, dat) {
    #                  -weibull.loglik(x, dat)
    #                },
    #                method = "BFGS",
    #                hessian = TRUE, dat = dat)
    #mles <- ml.fit$par
    w.value <- 2*(weibull.loglik(mles, dat = dat) -
                  weibull.loglik(beta0, dat = dat))
    diff1 <- as.matrix(mles - beta0)
    jhat1 <- weibull.info(mles, dat = dat)
    jhat1Inverse <- solve(jhat1)
    tSq.value <- t(diff1)%*%jhat1%*%diff1
    strue <- as.matrix(weibull.score(beta0, dat = dat))
    sSq.value <- t(strue)%*%jhat1Inverse%*%strue
    c(w.value, tSq.value, sSq.value, sum(abs(weibull.score(mles, dat = dat))))
  }
  weibull.samples <- rweibull(nsimu*ssize, shape = beta0[2],
                              scale = beta0[1]^(-1/beta0[2]))
  weibull.samples <- matrix(weibull.samples,
                            nrow = nsimu,
                            ncol = ssize)
  res <- apply(weibull.samples, 1, values1)
  if (any(res[4,] > 0.1)) stop("Some fits did not converge")
  w.cover <- sapply(conf.levels,
                    function(x) mean((res[1,] < qchisq(x, 2))))
  tSq.cover <- sapply(conf.levels,
                      function(x) mean((res[2,] < qchisq(x, 2))))
  sSq.cover <- sapply(conf.levels,
                      function(x) mean((res[3,] < qchisq(x, 2))))
  simu.se <- sqrt(conf.levels*(1-conf.levels)/nsimu)
  res <- cbind(conf.levels, w.cover, tSq.cover, sSq.cover, simu.se)
  colnames(res) <- c("nominal", "w", "t^2", "s^2", "simu s.e.")
  res
}

## Estimated coverages for ssize = 20, 50, 100
res20 <- weibull.simu(nsimu = 10000, ssize = 20, beta0 = c(1, 1))
res50 <- weibull.simu(nsimu = 10000, ssize = 50, beta0 = c(1, 1))
res100 <- weibull.simu(nsimu = 10000, ssize = 100, beta0 = c(1, 1))


###############################
### Change in parameterization
##############################
phis <- function(beta) log(beta)
betas <- function(phi) exp(phi)
m.weibull.loglik2 <- function(phi, dat)
  -weibull.loglik(betas(phi), dat)
dbetadphi <- function(phi) diag(exp(phi))
dev.new()
par(mfrow = c(1,2))
##
##
## Confidence regions based on w(lambda, nu)
##
llikvalues <- apply(parvalues, 1, weibull.loglik, dat = y)
llikvalues <- matrix(llikvalues,
                     nrow = length(lambdas),
                     ncol = length(nus))
contour(lambdas, nus, 2*(max(llikvalues) - llikvalues),
        levels = qchisq(conf.levels, 2),
        labels = as.character(conf.levels),
        xlab = expression(lambda),
        ylab = expression(nu),
        col = color1, lwd = lwds)
points(x = mles[1], y = mles[2], col = "red", pch =16)
title(expression(paste("Confidence regions based on ", w(lambda, nu),
    " at various levels")))

phi1s <- log(nus)
phi2s <-  log(lambdas)
contour(phi1s, phi2s, 2*(max(llikvalues) - llikvalues),
        levels = qchisq(conf.levels, 2),
        labels = as.character(conf.levels),
        xlab = expression(phi[1]),
        ylab = expression(phi[2]),
        col = color1, lwd = lwds, xlim = c(-1.2, 1.2), ylim = c(-1.2,1.2))
points(x = phis(mles[1]), y = phis(mles[2]), col = "red", pch =16)
title(expression(paste("Confidence regions based on ", w(phi[1], phi[2]),
    " at various levels")))

